/* Name: main.c
 * Author: <insert your name here>
 * Copyright: <insert your copyright message here>
 * License: <insert your license reference here>
 */

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>


// Special PC Node has serial code and no indicator LED's
// Collects scores and starts games.
// #define PC_NODE


#ifdef PC_NODE
#define output(directions,pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for input
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 100 // bit delay for 9600 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay

#define input_port PORTA
#define input_direction DDRA
#define input_pin (1 << PA1)
#define input_pins PINA
#define serial_port PORTA
#define serial_direction DDRA
#define serial_pin_out (1 << PA0)

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
    //
    // send character in txchar on port pin
    //    assumes line driver (inverts bits)
    //
    // start bit
    //
    clear(*port,pin);
    bit_delay();
    //
    // unrolled loop to write data bits
    //
    if bit_test(txchar,0)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,1)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,2)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,3)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,4)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,5)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,6)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    if bit_test(txchar,7)
        set(*port,pin);
    else
        clear(*port,pin);
    bit_delay();
    //
    // stop bit
    //
    set(*port,pin);
    bit_delay();
    //
    // char delay
    //
    bit_delay();
}

void put_string(volatile unsigned char *port, unsigned char pin, const unsigned char *data) {
    while( *data != '\0' )
    {
        put_char( port, pin, *data++ );
    }
}

unsigned char nibble_to_hex(unsigned char b)
{
    unsigned char ret = b & 0x0f;
    if(ret>9)
        ret += ('A'-10);
    else
        ret += '0';
    return ret;
}
#endif // PC_NODE


// Collision avoidance scheme:
// 1) Check channel is clear before sending
// 2) Re-transmit at slightly random intervals

// Bit encoding scheme:
// Was going to do a proper fixed bit width NRZ scheme as per RC5 spec, but I'm out of time and tired so I've simplified it to a variable pulse width scheme.
// 300us pulse = 0
// 600us pulse = 1
// 300us pause between bits
#define BIT_GAP     500 //300
#define BIT_START   1500 //900
#define BIT_0       500 //   300
#define BIT_1       1000 //600
#define BIT_TEST_S1 (1250/13) //(750/13)
#define BIT_TEST_10 (750/13) //(450/13)


// Byte encoding scheme:
// use an extra long 900us pulse as a start byte condition, only actually required at the start of 1st byte in message
// Send MSB first

// Most small messages will be 32 bits long
// At average 600us bit rate this will take approx 20ms to transmit

// Messages:
// Simple Packet Format - for now all messages will be 4 bytes as I'm running out of time to implement
// Byte 0 - Src addr : Dest addr
// Byte 1 - Length : Seq number
// Byte 2 - Function : payload
// Byte 3 - CSUM


// Comms Scheme:
// Sequence number starts from a random seed based on time to first button press?
// Ask for address, use seq number to distinguish from other address requests
// Roughly once per second send score to b'cast address
// When trigger pressed send 'shoot1' to b'cast address


// Addresses
// Address 0xF is braodcast address
#define NODE_ADDR 0b10101010
//#define NODE_ADDR 0x01


// Functions
#define FUNC_NONE           0
#define FUNC_SHOOT1         1
#define FUNC_SHOOT2         2
#define FUNC_SHOOT3         3
#define FUNC_REPORT_DEATH   4
#define FUNC_SEND_SCORE     5
#define FUNC_REQUEST_SCORE  6
#define FUNC_START_GAME     7
#define FUNC_END_GAME       8
#define FUNC_ALLOCATE_TEAM  9
//
#define FUNC_REQUEST_ADDR   14
#define FUNC_ALLOCATE_ADDR  15


// Trigger Button
#define TRIGGER_PIN         _BV(PA5)

// Buzzer Output
#define BUZZER_PIN          _BV(PA4)

// IR Tx
#define IRLED_CARRIER_PIN   _BV(PB2)
#define IRLED_OUTPUT_PIN    _BV(PA6) // Used to modulate the carrier
// IR Rx
#define IRRX_IN_PIN         _BV(PA7)


//
#define TIM1_TOP (1<<14) // 16384
#define TIM1_TOP_MSK (TIM1_TOP-1);
#define OLD_MESSAGE_TIMEOOUT   200  // about 2.5ms

//
// Game State Vars
//
// !!!! Need to check if operations on these volatiles are atomic !!!!!
volatile uint8_t        gLife = 3;
volatile uint32_t       gHitBlanking = 0;



enum eIRRXState
{
    eIDLE = 0,
    eINBIT,
    eBLANKING,
};


volatile uint32_t input_edge_last_count = 0;
uint32_t input_edge_start_count = 0;
//volatile uint8_t ir_rx_message = 0;
volatile uint8_t IRRX_State = eIDLE;

volatile uint8_t IR_RX_BitCount = 0;
volatile uint8_t IR_RX_Message = 0;

// Timer 1 OCR1A, used to generate application ticks at 4Hz
volatile uint32_t ticks = 0;
ISR(TIM1_COMPA_vect)
{
    ticks++;
    
    
    // Check and timeout any really old partial messages
    if(TIM1_TOP - input_edge_last_count > OLD_MESSAGE_TIMEOOUT)
    {
        if(IRRX_State != eBLANKING) // if we're blanking due to TX then just leave it
        {
            // Reset message rx handling
            TCCR1B &=  0b10111111; // ~ICES1;  // set to detect falling edge
            IRRX_State = eIDLE;
            IR_RX_Message = 0;
            IR_RX_BitCount = 0;
        }
    }
}

// It's late this needs a complete re-work...
// Timer 1 Input Capture, use to time pulse width of IR Rx signal
// ICP pin idles high and goes low when 38KHz carrier detected

ISR(TIM1_CAPT_vect)
{
    uint32_t count = ICR1;
    

    input_edge_last_count = count; // keep a copy to time out messages
    
    switch(IRRX_State)
    {
        case eIDLE: // start of bit, falling edge detected
            TCCR1B |= 0b01000000; //ICES1; // swap to detect rising edge
            input_edge_start_count = count;
            IRRX_State = eINBIT;
            break;
        case eINBIT: // end of bit, rising edge detected
            TCCR1B &= 0b10111111; //~ICES1;  // swap to detect falling edge
            count = (count - input_edge_start_count) & TIM1_TOP_MSK; // modulo math to deal with timer wrapping mid message
            if(count > BIT_TEST_S1) // 750us
            {
                // start condition
                IR_RX_Message = 0;
                IR_RX_BitCount = 0;
            }
            else if(count > BIT_TEST_10) // 450us mid VPW test point
            {
                // It's a 1
                IR_RX_Message = (IR_RX_Message<<1) | 0x01;
                IR_RX_BitCount++;
            }
            else
            {
                // It's a 0
                IR_RX_Message = (IR_RX_Message<<1);
                IR_RX_BitCount++;
            }
            IRRX_State = eIDLE;
            
            if(IR_RX_BitCount == 8)
            {
                // Add to message Q
                
                // Dirty game test
                // if top bits set hen we've got something
                if((IR_RX_Message & 0xc0) && (gLife > 0))
                //if(((IR_RX_Message & 0xf0) == FUNC_SHOOT1) && (gLife > 0))
                {
                    if(/*(gHitBlanking==0) ||*/ (ticks > gHitBlanking))
                    {
                        gLife--;
                        gHitBlanking = ticks + 8; // buzz and blank new hits for 2 seconds
                        PORTA |= BUZZER_PIN; // turn on buzzer
                    }
                }
                
                IR_RX_BitCount = 0;
                IR_RX_Message = 0;
            }
            
            break;
        case eBLANKING:
            break;
    }
}


// Simple random number generator
// Stolen from Neil's Speaker demo
uint16_t lfsr = 0b1010101000111000; // Can't init with 0!
uint8_t random(void)
{
    uint16_t bit;
    //
    // update 11 bit LFSR
    //
    bit  = ((lfsr >> 0) ^ (lfsr >> 9)) & 1;
    lfsr =  (lfsr >> 1) | (bit << 10);
    return lfsr&0xff;
}


// Send byte via IR
// Very poor for now, just delay
void IR_SendByte(uint8_t byte)
{
    uint8_t bit = byte & 0x80;
    uint8_t i;
    
    _delay_us(BIT_GAP); // pause
    
    // Check media clear
    // wait while IRRX_State not idle?
    
    // Force blanking state
    // Really dirty but we see our own signal realy loudly!
    IRRX_State = eBLANKING;
    
    // Send start condition
    PORTA |= IRLED_OUTPUT_PIN; // Set IR carrier on
    _delay_us(BIT_START);
    PORTA &= ~IRLED_OUTPUT_PIN; // Set IR carrier off
    _delay_us(BIT_GAP); // pause
    
    for(i=0; i<8; i++)
    {
        PORTA |= IRLED_OUTPUT_PIN; // Set IR carrier on
        
        if(bit != 0)
        {
            _delay_us(BIT_0);
        }
        else
        {
            _delay_us(BIT_1);
        }
        PORTA &= ~IRLED_OUTPUT_PIN; // Set IR carrier off
        _delay_us(BIT_GAP); // pause
        byte <<= 1;
        bit = byte & 0x80;
    }
    
    // Force back to idle state
    // Reset message rx handling
    TCCR1B &=  0b10111111; // ~ICES1;  // set to detect falling edge
    IRRX_State = eIDLE;
    IR_RX_Message = 0;
    IR_RX_BitCount = 0;
}


void shoot(void)
{
    //IR_SendByte((FUNC_SHOOT1<<4) | NODE_ADDR);
    IR_SendByte(NODE_ADDR);
}


int main(void)
{
    int i;
    uint8_t     temp = 0;
    uint32_t    last_ticks = 0;
    uint32_t    bBuzzerTime = 0; // tick count when buzzer due to go off
    uint32_t    bIRTime = 0;
    uint8_t     indicator_leds = 0;
    
    

    // Set timer 0 (8bit) to generate a 50% duty cycle 38KHz IR carrier on OC0A (PB2)
    // Could use one of the proper PWN modes to make it 25% duty cycle as per IR5 spec
    // Requires OC0A timeout at 76KHz to toggle pin
    // Set CTC mode with OC0A toggling on each match
    TCCR0A = 0b01000010;
    // No prescale
    TCCR0B = 0b00000001;
    TCNT0 = 0;
    OCR0A = 255; //F_CPU/76000; was 211
    
    // Set timer 1 (16bit) to time input capture signal edges
    // and to generate application ticks
    // Must be able to distinguish 300us from 600us timing between edges => at least 2 counts per 300us (or count > 6.7KHz)
    // Must have resolution not to overflow more than once during message reception
    // Application ticks of 50Hz or 100Hz would be nice.  4Hz is OK.
    // Set CTC mode with OC1A IRQ to update application ticks, OC1A pin not enabled
    TCCR1A = 0b00000000;
    // Enable ICP noise canceller, initially trigger on negative edge (start condition)
    // CTC Mode, prescale 1/256 (78.125KHz at 20MHz fosc)
    TCCR1B= 0b10001100;
    TCNT1H = 0;
    TCNT1L = 0;
    // Set OCR1A IRQ at 4Hz (19531), but actually make it nearest power of 2 (16384) to make our modulo math for edge timing simpler = 4.7Hz
    OCR1A = 16384; //TIM1_TOP; //16384;
    //OCR1AH = 0x4c;
    //OCR1AL = 0x4b;
    // Enable OCR1A and ICP IRQs
    TIMSK1 = 0b00100010;
    
    //
    // Damn! Didn't notice during HW design but these 2 timers are basically running at the same rate, should just combine them into one.
    //
    
    
    // Set up IR LED
    DDRB |= IRLED_CARRIER_PIN; // Set OC0A as an output
    DDRA |= IRLED_OUTPUT_PIN; // Set LED on/off pin as an output
    
    // Set up IR RX
    DDRA &= ~IRRX_IN_PIN; // Set IR RX pin as an input (should boot up this way)
    PORTA |= IRRX_IN_PIN; // Enable pull up
    
    // Set up indicator LEDS
    DDRA |= 0x0f; // Set all LED indication pins as outputs
    
    // Set up Trigger
    DDRA &= ~TRIGGER_PIN; // Set IR RX pin as an input (should boot up this way)
    PORTA |= TRIGGER_PIN; // Enable pull up

    // Set up Buzzer
    DDRA |= BUZZER_PIN;
    
    sei(); // Enable irq's
    
    
    for(;;)
    {
    
#if 0
        if( ticks != last_ticks)
        {
            last_ticks = ticks;
            PORTA &= ~0x0f;
            PORTA |= (random()&0x0f);
        }
#endif
 
#if 1
        // Handle buzzer
        if(bBuzzerTime != 0)
        {
            if(ticks >= bBuzzerTime)
            {
                // turn buzzer off
                PORTA &= ~BUZZER_PIN;
                bBuzzerTime = 0;
            }
        }
        
        // Handle trigger
        if((PINA & TRIGGER_PIN) == 0) // trigger pressed
        {
            if(bIRTime !=0)
            {
                if(ticks >= bIRTime)
                {
                    if( gLife > 0)
                    {
                        PORTA |= BUZZER_PIN; // turn on buzzer
                        bBuzzerTime = ticks + 1; // turn buzzer off after 1 tick
                        bIRTime = ticks + 2; // send message repeat after 2 ticks
                        //IR_SendByte(NODE_ADDR);
                        shoot();
                    }
                }
            
            }
            else
            {
                // New trigger press
                if(gLife > 0)
                {
                    PORTA |= BUZZER_PIN; // turn on buzzer
                    bBuzzerTime = ticks + 1; // turn buzzer off after 1 tick
                    bIRTime = ticks + 2; // send message repeat after 2 ticks
                    //IR_SendByte(NODE_ADDR);
                    shoot();
                }
            }
        }
        else
            bIRTime = 0; // no trigger so stop sending kill message
        
        // Handle hit blanking
        if(gHitBlanking != 0)
        {
            bBuzzerTime = gHitBlanking;
            if(ticks>gHitBlanking)
                gHitBlanking = 0;
        }
        
        if((ticks - last_ticks) >= 2)
        {
            // Do some processing
            
            last_ticks = ticks;
        }
        
        // Force IR RX low nibble display
        //temp = ~IR_RX_Message; // LED's are inverted
        
        // Set LEDs to show life
        if(gLife > 0)
        {
            //temp = ((~gLife) << 1) | 1;
            temp = 0x07>>(3-gLife);
            temp <<= 1;
            temp = ~temp; // LED's are inverted
        }
        else
            temp = 0b00001110;
        
        // Display LEDs
        if(temp != indicator_leds) // Only display if changed
        {
            PORTA &= ~0x0f;
            PORTA |= (temp&0x0f);
            indicator_leds = temp;
        }
#endif
        
        // Display current health
        
        
        // If dead then skip rest of game play
        
        
        // If button pressed then shoot, inc shots fired count, do short buzz
        
        
        // If hot then adjust score, do long buzz
        
        
        // Report my score via IR
        
        
        // Report all scores over serial
        // Shots fired
        // Shot by addr0
        // shot by addr1
        // ...
        
    }
    return 0;   /* never reached */
}
